// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "Tileset.hpp"
#include "assert/assert.hpp"

using namespace Dawn;

struct Tile Tileset::getTile(const int32_t tile) {
  assertTrue(tile >= 0, "Tile must be greater than or equal to 0");
  assertTrue(tile < this->tiles.size(), "Tile is out of bounds");
  return this->tiles[tile];
}

TilesetGrid::TilesetGrid() {

}

TilesetGrid::TilesetGrid(
  Texture &texture,
  const int32_t columns,
  const int32_t rows
) : TilesetGrid(
  columns, rows,
  texture.getWidth(), texture.getHeight(),
  0, 0,
  0, 0
) {

}

TilesetGrid::TilesetGrid(
  const int32_t columns,
  const int32_t rows,
  const int32_t w,
  const int32_t h,
  const int32_t gapX,
  const int32_t gapY,
  const int32_t borderX,
  const int32_t borderY
) {
  assertTrue(columns >= 1, "Columns must be greater than or equal to 1");
  assertTrue(rows >= 1, "Rows must be greater than or equal to 1");
  assertTrue(w >= 1, "Width must be greater than or equal to 1");
  assertTrue(h >= 1, "Height must be greater than or equal to 1");
  assertTrue(gapX >= 0, "GapX must be greater than or equal to 0");
  assertTrue(gapY >= 0, "GapY must be greater than or equal to 0");
  assertTrue(borderX >= 0, "BorderX must be greater than or equal to 0");
  assertTrue(borderY >= 0, "BorderY must be greater than or equal to 0");
  assertTrue(w >= (columns + (gapX * columns) + borderX + borderX), "Width is too small");
  assertTrue(h >= (rows + (gapY * rows) + borderY + borderY), "Height is too small");

  this->rows = rows;
  this->columns = columns;

  // Calculate division sizes (pixels)
  this->divX = (w - (borderX * 2) - (gapX * (columns - 1))) / columns;
  this->divY = (h - (borderY * 2) - (gapY * (rows - 1))) / rows;

  // Calculate the division sizes (units)
  const float_t tdivX = (float_t)this->divX / (float_t)w;
  const float_t tdivY = (float_t)this->divY / (float_t)h;

  struct Tile tile;
  for(int32_t y = 0; y < rows; y++) {
    for(int32_t x = 0; x < columns; x++) {
      tile.uv0.x = (borderX + ((float_t)this->divX * x) + (gapX * x)) / w;
      tile.uv1.x = tile.uv0.x + tdivX;
          
      tile.uv0.y = (borderY + ((float_t)this->divY * y) + (gapY * y)) / h;
      tile.uv1.y = tile.uv0.y + tdivY;
      this->tiles.push_back(tile);
    }
  }
}

float_t TilesetGrid::getTileWidth(const int32_t tile) {
  return this->divX;
}

float_t TilesetGrid::getTileHeight(const int32_t tile) {
  return this->divY;
}

struct Tile TilesetGrid::getTileFromGrid(
  const int32_t column,
  const int32_t row
) {
  assertTrue(row > 0 && row < this->rows, "Row is out of bounds");
  assertTrue(column > 0 && column < this->columns, "Column is out of bounds");
  return this->getTile(row + (column * this->rows));
}